/*
 * (C) 2002 Antonino Perricone <don_perricone@libero.it>
 * (C) 2013 Ohtsuka Yoshio <ohtsuka.yoshio@gmail.com>
 *
 * This work is licensed under a Creative Commons 
 * Attribution-NonCommercial-ShareAlike 3.0 Unported License.
 * See this link for more informations: 
 * http://creativecommons.org/licenses/by-nc-sa/3.0/
 */
#include <glib.h>
#include <gtk/gtk.h>
#include "luce_gimp_main.h"
#include "luce_gimp_dialog.h"
#include "luce_gimp_preview.h"

#ifdef _WIN32
#ifdef DATADIR
#undef DATADIR
#endif
#include <windows.h>
#else
#include <unistd.h>
#endif

#define LUCE_GIMP_RESPONSE_OK     0
#define LUCE_GIMP_RESPONSE_CANCEL 1


typedef struct _DialogWidgets {
	PluginError		dialog_result;
	GimpDrawable	*drawable;
	GtkWidget		*dialog;
	GtkWidget		*preview;

	GtkWidget		*light_type_point;
	GtkWidget		*light_type_directional;
	GtkWidget		*light_type_quadric_attenuation;

	GtkWidget		*light_type_sub_options;

	GtkWidget		*center_x;
	GtkWidget		*center_y;
	GtkWidget		*unit;
	GimpUnit		old_unit_value;

	GtkWidget		*angle;

	GtkWidget		*algorithm_type_zero;
	GtkWidget		*algorithm_type_light_form;
	GtkWidget		*algorithm_type_option_table;

	GtkWidget		*color;
	GtkWidget		*mask;
	GtkWidget		*normalize_light;
	GtkWidget		*normalize_shadow;

	GtkWidget		*scale_light;
	GtkWidget		*scale_light_slider;
	GtkObject		*adjustment_light_scale;

	GtkWidget		*scale_shadow;
	GtkWidget		*scale_shadow_slider;
	GtkObject		*adjustment_shadow_scale;

	GtkWidget		*source_add;
	GtkWidget		*source_zero;
	GtkWidget		*source_normalize;

	GtkWidget		*thread_number;
} DialogWidgets;

static DialogWidgets widgets;

#define INTENSITY_SLIDER_POS_TO_SPIN_VALUE(POS) (pow (10, (POS) / 1000.0 - 2.0) * 100.0)
#define INTENSITY_SPIN_VALUE_TO_SLIDER_POS(VAL) ((log10 ((VAL) * 100.0) - 2.0) * 1000.0)
#define INTENSITY_SPIN_VALUE_TO_LUCE_VALUE(VAL) ((VAL) / 100.0)
#define INTENSITY_LUCE_VALUE_TO_SPIN_VALUE(VAL) ((VAL) * 100.0)


static void response_callback								(GtkWidget		*dialog,
															 gint			response_id,
															 LuceOptions	*luce_options);
static void light_type_point_clicked_callback				(GtkWidget		*widget,
															 LuceOptions	*luce_options);
static void light_type_directional_clicked_callback			(GtkWidget		*widget,
															 LuceOptions	*luce_options);
static void light_type_quadric_attenuation_clicked_callback	(GtkWidget		*widget,
															 LuceOptions	*luce_options);
static void update_center_position_unit						();
static void center_x_value_changed_callback					(GtkWidget		*widget,
															 LuceOptions	*luce_options);
static void center_y_value_changed_callback					(GtkWidget		*widget,
															 LuceOptions	*luce_options);
static void unit_changed_callback							(GtkWidget		*widget,
															 LuceOptions	*luce_options);
static void angle_value_changed_callback					(GtkWidget		*widget,
															 LuceOptions	*luce_options);
static void algorithm_type_zero_clicked_callback			(GtkWidget		*widget,
															 LuceOptions	*luce_options);
static void algorithm_type_light_form_clicked_callback		(GtkWidget		*widget,
															 LuceOptions	*luce_options);
static void color_color_changed_callback					(GtkWidget		*widget,
															 LuceOptions	*luce_options);
static void mask_value_changed_callback						(GtkWidget		*widget,
															 LuceOptions	*luce_options);
static void normalize_light_clicked_callback				(GtkWidget		*widget,
															 LuceOptions	*luce_options);
static void normalize_shadow_clicked_callback				(GtkWidget		*widget,
															 LuceOptions	*luce_options);
static void scale_light_value_changed_callback				(GtkWidget		*widget,
															 LuceOptions	*luce_options);
static void scale_shadow_value_changed_callback				(GtkWidget		*widget,
															 LuceOptions	*luce_options);
static void adjustment_light_scale_value_changed			(GtkAdjustment	*object,
															 LuceOptions	*luce_options);
static void adjustment_shadow_scale_value_changed			(GtkAdjustment	*object,
															 LuceOptions	*luce_options);
static void update_source_sensitive							();
static void source_add_clicked_callback						(GtkWidget		*widget,
															 LuceOptions	*luce_options);
static void source_zero_clicked_callback					(GtkWidget		*widget,
															 LuceOptions	*luce_options);
static void source_normalize_clicked_callback				(GtkWidget		*widget,
															 LuceOptions	*luce_options);
static void thread_number_changed_callback					(GtkWidget		*widget,
															 LuceOptions	*luce_options);
static int	get_processor_number							();

static void
response_callback (GtkWidget	*dialog,
				   gint			response_id,
				   LuceOptions	*luce_options)
{
	switch (response_id) {
		case LUCE_GIMP_RESPONSE_OK:
			widgets.dialog_result = ERROR_NOTHING;
			gtk_widget_destroy(dialog);
			break;

		case LUCE_GIMP_RESPONSE_CANCEL:
			widgets.dialog_result = ERROR_CANCEL;
			gtk_widget_destroy(dialog);
			break;
	}
}

static void
light_type_point_clicked_callback (GtkWidget	*widget,
								   LuceOptions	*luce_options)
{
	luce_options->SetPointLight();
	gtk_notebook_set_current_page (GTK_NOTEBOOK (widgets.light_type_sub_options), 0);

	luce_gimp_preview_update_buffer (luce_options);
	gtk_widget_queue_draw (widgets.preview);
}

static void 
light_type_directional_clicked_callback (GtkWidget		*widget,
										 LuceOptions	*luce_options)
{
	luce_options->SetDirectionaLight();
	gtk_notebook_set_current_page (GTK_NOTEBOOK (widgets.light_type_sub_options), 1);

	luce_gimp_preview_update_buffer (luce_options);
	gtk_widget_queue_draw (widgets.preview);
}

static void
light_type_quadric_attenuation_clicked_callback (GtkWidget		*widget,
												 LuceOptions	*luce_options)
{
	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget))) {
		luce_options->SetQuadraticAttenuation();
	} else {
		luce_options->SetLinearAttenuation();
	}

	luce_gimp_preview_update_buffer (luce_options);
	gtk_widget_queue_draw (widgets.preview);
}

static void
update_center_position_unit ()
{
	GimpUnit old_unit = widgets.old_unit_value;
	GimpUnit new_unit = gimp_unit_combo_box_get_active (GIMP_UNIT_COMBO_BOX (widgets.unit));
	int image_ID = gimp_drawable_get_image (widgets.drawable->drawable_id);
	gdouble x = gtk_spin_button_get_value (GTK_SPIN_BUTTON (widgets.center_x));
	gdouble y = gtk_spin_button_get_value (GTK_SPIN_BUTTON (widgets.center_y));
	gdouble resx, resy;
	gdouble px, py;

	if (new_unit == GIMP_UNIT_PIXEL) {
		gtk_spin_button_set_digits (GTK_SPIN_BUTTON (widgets.center_x), 0);
		gtk_spin_button_set_digits (GTK_SPIN_BUTTON (widgets.center_y), 0);
	} else {
		gtk_spin_button_set_digits (GTK_SPIN_BUTTON (widgets.center_x), 3);
		gtk_spin_button_set_digits (GTK_SPIN_BUTTON (widgets.center_y), 3);
	}

	gimp_image_get_resolution (image_ID, &resx, &resy);

	if (old_unit == GIMP_UNIT_PERCENT) {
		px = widgets.drawable->width  * x / 100.0;
		py = widgets.drawable->height * y / 100.0;
	} else {
		px = gimp_units_to_pixels (x, old_unit, resx);
		py = gimp_units_to_pixels (y, old_unit, resy);
	}

	if (new_unit == GIMP_UNIT_PERCENT) {
		x = px / widgets.drawable->width  * 100.0;
		y = py / widgets.drawable->height * 100.0;
	} else {
		x = gimp_pixels_to_units (px, new_unit, resx);
		y = gimp_pixels_to_units (py, new_unit, resy);
	}

	gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.center_x), x);
	gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.center_y), y);

	widgets.old_unit_value = new_unit;
}

static void 
center_x_value_changed_callback (GtkWidget		*widget,
								 LuceOptions	*luce_options)
{
	gdouble value = gtk_spin_button_get_value (GTK_SPIN_BUTTON (widget));
	GimpUnit unit;
	gint image_ID;
	gdouble resx, resy;
	
	image_ID = gimp_drawable_get_image (widgets.drawable->drawable_id);
	gimp_image_get_resolution (image_ID, &resx, &resy);
	unit = gimp_unit_combo_box_get_active (GIMP_UNIT_COMBO_BOX (widgets.unit));

	if (unit == GIMP_UNIT_PERCENT) {
		value /= 100.0;
	} else {
		value = gimp_units_to_pixels (value, unit, resx) 
			/ gimp_drawable_width (widgets.drawable->drawable_id);
	}
	luce_options->SetCenter(
		value,
		luce_options->GetCenter().y);

	luce_gimp_preview_update_buffer (luce_options);
	gtk_widget_queue_draw (widgets.preview);
}

static void 
center_y_value_changed_callback (GtkWidget		*widget,
								 LuceOptions	*luce_options)
{
	gdouble value = gtk_spin_button_get_value (GTK_SPIN_BUTTON (widget));
	GimpUnit unit;
	gint image_ID;
	gdouble resx, resy;
	
	image_ID = gimp_drawable_get_image (widgets.drawable->drawable_id);
	gimp_image_get_resolution (image_ID, &resx, &resy);
	unit = gimp_unit_combo_box_get_active (GIMP_UNIT_COMBO_BOX (widgets.unit));

	if (unit == GIMP_UNIT_PERCENT) {
		value /= 100.0;
	} else {
		value = gimp_units_to_pixels (value, unit, resy) 
			/ gimp_drawable_height (widgets.drawable->drawable_id);
	}
	luce_options->SetCenter(luce_options->GetCenter().x, value);

	luce_gimp_preview_update_buffer (luce_options);
	gtk_widget_queue_draw (widgets.preview);
}

static void
unit_changed_callback (GtkWidget	*widget,
					   LuceOptions	*luce_options)
{
	update_center_position_unit ();
}

static void
angle_value_changed_callback (GtkWidget		*widget,
							  LuceOptions	*luce_options)
{
	gdouble angle = gtk_spin_button_get_value (GTK_SPIN_BUTTON (widget)) * G_PI / 180.0;
	luce_options->SetDirection (-(float)cos(angle), (float)sin(angle));

	luce_gimp_preview_update_buffer (luce_options);
	gtk_widget_queue_draw (widgets.preview);
}

static void
algorithm_type_zero_clicked_callback (GtkWidget		*widget,
									  LuceOptions	*luce_options)
{
	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget))) {
		luce_options->SetZeroMode();
	}
	gtk_widget_set_sensitive (widgets.algorithm_type_option_table, TRUE);

	luce_gimp_preview_update_buffer (luce_options);
	gtk_widget_queue_draw (widgets.preview);
}

static void
algorithm_type_light_form_clicked_callback (GtkWidget	*widget,
											LuceOptions	*luce_options)
{
	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget))) {
		luce_options->SetLightformMode();
	}
	gtk_widget_set_sensitive (widgets.algorithm_type_option_table, FALSE);

	luce_gimp_preview_update_buffer (luce_options);
	gtk_widget_queue_draw (widgets.preview);
}

static void
color_color_changed_callback (GtkWidget		*widget,
							  LuceOptions	*luce_options)
{
	GimpRGB color;
	gimp_color_button_get_color (GIMP_COLOR_BUTTON (widget), &color);
	luce_options->SetZero(0, (float)color.r);
	luce_options->SetZero(1, (float)color.g);
	luce_options->SetZero(2, (float)color.b);

	luce_gimp_preview_update_buffer (luce_options);
	gtk_widget_queue_draw (widgets.preview);
}

static void
mask_value_changed_callback (GtkWidget		*widget,
							 LuceOptions	*luce_options)
{
	gdouble value = gtk_spin_button_get_value (GTK_SPIN_BUTTON (widget));
	luce_options->SetZero(4, (float)value);

	luce_gimp_preview_update_buffer (luce_options);
	gtk_widget_queue_draw (widgets.preview);
}

static void
normalize_light_clicked_callback (GtkWidget		*widget,
								  LuceOptions	*luce_options)
{
	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget))) {
		luce_options->SetNormalizePositive();
	} else {
		luce_options->SetMantainPositiveScale();
	}

	luce_gimp_preview_update_buffer (luce_options);
	gtk_widget_queue_draw (widgets.preview);
}

static void
normalize_shadow_clicked_callback (GtkWidget	*widget,
								   LuceOptions	*luce_options)
{	
	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget))) {
		luce_options->SetNormalizeNegative();
	} else {
		luce_options->SetMantainNegativeScale();
	}

	luce_gimp_preview_update_buffer (luce_options);
	gtk_widget_queue_draw (widgets.preview);
}

static void
scale_light_value_changed_callback (GtkWidget	*widget,
									LuceOptions	*luce_options)
{
	gdouble value;
	value = gtk_spin_button_get_value (GTK_SPIN_BUTTON (widget));
	luce_options->SetPositiveIntensity(
		(float)INTENSITY_SPIN_VALUE_TO_LUCE_VALUE (value));
	gtk_adjustment_set_value (GTK_ADJUSTMENT (widgets.adjustment_light_scale),
							  INTENSITY_SPIN_VALUE_TO_SLIDER_POS (value));

	luce_gimp_preview_update_buffer (luce_options);
	gtk_widget_queue_draw (widgets.preview);
}

static void
scale_shadow_value_changed_callback (GtkWidget		*widget,
									 LuceOptions	*luce_options)
{
	gdouble value;
	value = gtk_spin_button_get_value (GTK_SPIN_BUTTON (widget));
	luce_options->SetNegativeIntensity(
		(float)INTENSITY_SPIN_VALUE_TO_LUCE_VALUE (value));
	gtk_adjustment_set_value (GTK_ADJUSTMENT (widgets.adjustment_shadow_scale),
							  INTENSITY_SPIN_VALUE_TO_SLIDER_POS (value));

	luce_gimp_preview_update_buffer (luce_options);
	gtk_widget_queue_draw (widgets.preview);
}

static void
adjustment_light_scale_value_changed (GtkAdjustment	*adjustment,
									  LuceOptions	*luce_options)
{
	gint slider_pos;

	slider_pos = (gint)gtk_adjustment_get_value (adjustment);
	gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.scale_light),
							   INTENSITY_SLIDER_POS_TO_SPIN_VALUE (slider_pos));
}

static void
adjustment_shadow_scale_value_changed (GtkAdjustment	*adjustment,
									   LuceOptions		*luce_options)
{
	gint slider_pos;

	slider_pos = (gint)gtk_adjustment_get_value (adjustment);
	gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.scale_shadow),
							   INTENSITY_SLIDER_POS_TO_SPIN_VALUE (slider_pos));
}

static void 
update_source_sensitive ()
{
	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widgets.source_add))) {
		gtk_widget_set_sensitive (widgets.source_zero, TRUE);
		if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widgets.source_zero))) {
			gtk_widget_set_sensitive (widgets.source_normalize, TRUE);
		} else {
			gtk_widget_set_sensitive (widgets.source_normalize, FALSE);
		}
	} else {
		gtk_widget_set_sensitive (widgets.source_zero, FALSE);
		gtk_widget_set_sensitive (widgets.source_normalize, FALSE);
	}
}
static void
source_add_clicked_callback (GtkWidget		*widget,
							 LuceOptions	*luce_options)
{
	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget))) {
		luce_options->SetAddSource();
	} else {
		luce_options->SetNotAddSource();
	}
	update_source_sensitive ();

	luce_gimp_preview_update_buffer (luce_options);
	gtk_widget_queue_draw (widgets.preview);
}

static void
source_zero_clicked_callback (GtkWidget		*widget,
							  LuceOptions	*luce_options)
{
	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget))) {
		luce_options->SetSrcApplyZero();
	} else {
		luce_options->SetSrcIgnoreZero();
	}
	update_source_sensitive ();

	luce_gimp_preview_update_buffer (luce_options);
	gtk_widget_queue_draw (widgets.preview);
}

static void
source_normalize_clicked_callback (GtkWidget	*widget,
								   LuceOptions	*luce_options)
{
	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget))) {
		luce_options->SetSrcNormalizePositive();
		luce_options->SetSrcNormalizeNegative();
	} else {
		luce_options->SetSrcMantainPositiveScale();
		luce_options->SetSrcMantainNegativeScale();
	}

	luce_gimp_preview_update_buffer (luce_options);
	gtk_widget_queue_draw (widgets.preview);
}

static void
thread_number_changed_callback (GtkWidget		*widget,
								LuceOptions	*luce_options)
{
	gint n;
	gimp_int_combo_box_get_active (GIMP_INT_COMBO_BOX (widget), &n);
	luce_options->SetNumThreadsToUse(n);

	luce_gimp_preview_update_buffer (luce_options);
	gtk_widget_queue_draw (widgets.preview);
}

static int
get_processor_number ()
{
	int num = 1;
#ifdef _WIN32
#ifndef _SC_NPROCESSORS_ONLN
SYSTEM_INFO info;
GetSystemInfo (&info);
#define sysconf(a) info.dwNumberOfProcessors
#define _SC_NPROCESSORS_ONLN
#endif
#endif

#ifdef _SC_NPROCESSORS_ONLN
	num = sysconf(_SC_NPROCESSORS_ONLN);
#endif
	return num;
}

void
luce_gimp_dialog_set_position (LuceOptions *luce_options, 
							   gdouble x,
							   gdouble y)
{
	gint image_ID;
	gdouble w, h;
	gdouble resx, resy;
	GimpUnit unit;

	image_ID = gimp_drawable_get_image (widgets.drawable->drawable_id);
	w = widgets.drawable->width;
	h = widgets.drawable->height;
	gimp_image_get_resolution (image_ID, &resx, &resy);
	unit = gimp_unit_combo_box_get_active (GIMP_UNIT_COMBO_BOX (widgets.unit));

	luce_gimp_preview_enable_update (false);

	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widgets.light_type_point))) {
		if (unit == GIMP_UNIT_PERCENT) {
			gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.center_x), x * 100.0);
			gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.center_y), y * 100.0);
		} else {
			gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.center_x), 
									   gimp_pixels_to_units (w * x, unit, resx));
			gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.center_y), 
									   gimp_pixels_to_units (h * y, unit, resy));
		}
	}

	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widgets.light_type_directional))) {
		gdouble x2, y2;
		x2 = w * x - w / 2;
		y2 = h * y - h / 2;
		gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.angle), 
								   atan2 (y2, -x2) / G_PI * 180);
	}

	luce_gimp_preview_enable_update (true);

	luce_gimp_preview_update_buffer (luce_options);
	gtk_widget_queue_draw (widgets.preview);
}

PluginError 
luce_gimp_dialog_run (GimpDrawable	*drawable,
					  LuceOptions	*luce_options) 
{
#define PADDING 2
	GtkWidget *main_hbox;
	GtkWidget *options_table;
	GtkWidget *light_option_frame;
	GtkWidget *algorithm_option_frame;
	GtkWidget *scale_option_frame;
	GtkWidget *source_option_frame;
	GtkWidget *thread_option_frame;

	gimp_ui_init (PLUGIN_NAME, TRUE);

	widgets.drawable = drawable;
	widgets.dialog = gtk_dialog_new ();
	gtk_window_set_title (GTK_WINDOW (widgets.dialog),
						  PLUGIN_NAME " Version " PLUGIN_VERSION);
	gtk_window_set_icon (GTK_WINDOW (widgets.dialog),
						 gdk_pixbuf_new_from_inline (-1, img_window_icon, FALSE, NULL));
	gtk_window_set_role (GTK_WINDOW (widgets.dialog), PLUGIN_NAME);
	gtk_window_set_resizable (GTK_WINDOW (widgets.dialog), FALSE);
	//gtk_window_set_keep_above (GTK_WINDOW (widgets.dialog), TRUE);
	gtk_dialog_set_has_separator (GTK_DIALOG (widgets.dialog), FALSE);

	gtk_dialog_add_buttons (GTK_DIALOG (widgets.dialog),
#ifdef _WIN32
							_("OK"),     LUCE_GIMP_RESPONSE_OK,
							_("Cancel"), LUCE_GIMP_RESPONSE_CANCEL,
#else
							_("Cancel"), LUCE_GIMP_RESPONSE_CANCEL,
							_("OK"),     LUCE_GIMP_RESPONSE_OK,
#endif
							NULL);
	g_signal_connect (widgets.dialog, "response", G_CALLBACK (response_callback), luce_options);
	g_signal_connect (widgets.dialog, "destroy", G_CALLBACK (gtk_main_quit), NULL);

	// Main hbox
	main_hbox = gtk_hbox_new (FALSE, 2);
	gtk_container_set_border_width (GTK_CONTAINER (main_hbox), 2);
	gtk_container_add (GTK_CONTAINER (GTK_DIALOG (widgets.dialog)->vbox), main_hbox);

	// Preview
	widgets.preview = luce_gimp_preview_new (luce_options, drawable);
	gtk_box_pack_start (GTK_BOX (main_hbox), widgets.preview, TRUE, TRUE, 0);

	// Options container table
	options_table = gtk_table_new (5, 2, false);
	gtk_box_pack_start (GTK_BOX (main_hbox), options_table, FALSE, FALSE, 2);

	// Light type
	light_option_frame = gtk_frame_new (_("Light Options"));
	gtk_table_attach (GTK_TABLE (options_table), light_option_frame, 0, 2, 0, 1, GTK_FILL, GTK_FILL, PADDING, PADDING);
	{
		GtkWidget *hbox = gtk_hbox_new (FALSE, PADDING);
		GtkWidget *vbox1 = gtk_vbox_new (FALSE, PADDING);
		GtkWidget *table1; //point type options
		GtkWidget *table2; //directionan type options
		GtkWidget *label_x;
		GtkWidget *label_y;
		GtkWidget *label_unit;
		GtkWidget *label_angle;

		widgets.light_type_sub_options = gtk_notebook_new ();

		gtk_container_add (GTK_CONTAINER (light_option_frame), hbox);
		gtk_container_add (GTK_CONTAINER (hbox), vbox1);
		gtk_container_add (GTK_CONTAINER (hbox), gtk_vseparator_new());
		gtk_container_add (GTK_CONTAINER (hbox), widgets.light_type_sub_options);
		gtk_container_set_border_width (GTK_CONTAINER (hbox), PADDING);
		gtk_container_set_border_width (GTK_CONTAINER (vbox1), PADDING);
		gtk_container_set_border_width (GTK_CONTAINER (widgets.light_type_sub_options), PADDING);

		gtk_notebook_set_show_border (GTK_NOTEBOOK (widgets.light_type_sub_options), FALSE);
		gtk_notebook_set_show_tabs (GTK_NOTEBOOK (widgets.light_type_sub_options), FALSE);

		// Light type Point
		widgets.light_type_point = gtk_radio_button_new_with_label (NULL, _("Point"));
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widgets.light_type_point),
									  luce_options->IsPointLight());
		g_signal_connect (widgets.light_type_point, 
						  "clicked",
						  G_CALLBACK (light_type_point_clicked_callback),
						  luce_options);
		gtk_box_pack_start (GTK_BOX (vbox1), widgets.light_type_point, FALSE, FALSE, PADDING);

		// Light type Directional
		widgets.light_type_directional = gtk_radio_button_new_with_label ( 
			gtk_radio_button_get_group (GTK_RADIO_BUTTON (widgets.light_type_point)),
			_("Directional"));
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widgets.light_type_directional),
									  luce_options->IsDirectionaLight());
		g_signal_connect (widgets.light_type_directional, 
						  "clicked",
						  G_CALLBACK (light_type_directional_clicked_callback),
						  luce_options);
		gtk_box_pack_start (GTK_BOX (vbox1), widgets.light_type_directional, FALSE, FALSE, PADDING);

		// Light type Quadric Attenuation
		widgets.light_type_quadric_attenuation = gtk_check_button_new_with_label (_("Quadric Attenuation"));
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widgets.light_type_quadric_attenuation),
									  luce_options->IsQuadraticAttenuation());
		g_signal_connect (widgets.light_type_quadric_attenuation,
						  "clicked",
						  G_CALLBACK (light_type_quadric_attenuation_clicked_callback),
						  luce_options);
		gtk_box_pack_start (GTK_BOX (vbox1), widgets.light_type_quadric_attenuation, FALSE, FALSE, PADDING);

		// Labels
		label_x = gtk_label_new ("X:");
		label_y = gtk_label_new ("Y:");
		label_unit = gtk_label_new (_("Unit:"));
		label_angle = gtk_label_new (_("Angle:"));
		gtk_misc_set_alignment (GTK_MISC (label_x), 0.0, 0.5);
		gtk_misc_set_alignment (GTK_MISC (label_y), 0.0, 0.5);
		gtk_misc_set_alignment (GTK_MISC (label_unit), 0.0, 0.5);
		gtk_misc_set_alignment (GTK_MISC (label_angle), 0.0, 0.5);

		// Light type Point Options
		table1 = gtk_table_new(3, 3, FALSE);
		gtk_notebook_append_page (GTK_NOTEBOOK (widgets.light_type_sub_options), table1, NULL);

		// Center X
		widgets.center_x = gtk_spin_button_new_with_range (SHRT_MIN, SHRT_MAX, 1);
		gtk_spin_button_set_digits (GTK_SPIN_BUTTON (widgets.center_x), 0);
		gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.center_x), 
								   drawable->width * luce_options->GetCenter().x);
		g_signal_connect (widgets.center_x,
						  "value-changed",
						  G_CALLBACK (center_x_value_changed_callback),
						  luce_options);

		// Center Y
		widgets.center_y = gtk_spin_button_new_with_range (SHRT_MIN, SHRT_MAX, 1);
		gtk_spin_button_set_digits (GTK_SPIN_BUTTON (widgets.center_y), 0);
		gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.center_y), 
								   drawable->height * luce_options->GetCenter().y);
		g_signal_connect (widgets.center_y,
						  "value-changed",
						  G_CALLBACK (center_y_value_changed_callback),
						  luce_options);

		// Unit
		{
			GimpUnitStore *unitstore = gimp_unit_store_new(10);
			gimp_unit_store_set_has_pixels (unitstore, TRUE);
			gimp_unit_store_set_has_percent (unitstore, TRUE);
			widgets.unit = gimp_unit_combo_box_new_with_model(unitstore);
			gimp_unit_combo_box_set_active (GIMP_UNIT_COMBO_BOX (widgets.unit),
											GIMP_UNIT_PIXEL);
			widgets.old_unit_value = GIMP_UNIT_PIXEL;
			g_signal_connect (widgets.unit,
							  "changed",
							  G_CALLBACK (unit_changed_callback),
							  luce_options);
		}

		gtk_table_attach (GTK_TABLE (table1), label_x,          0, 1, 0, 1, GTK_FILL, GTK_FILL, PADDING, PADDING);
		gtk_table_attach (GTK_TABLE (table1), widgets.center_x, 1, 2, 0, 1, GTK_FILL, GTK_FILL, PADDING, PADDING);
		gtk_table_attach (GTK_TABLE (table1), label_y,          0, 1, 1, 2, GTK_FILL, GTK_FILL, PADDING, PADDING);
		gtk_table_attach (GTK_TABLE (table1), widgets.center_y, 1, 2, 1, 2, GTK_FILL, GTK_FILL, PADDING, PADDING);
		gtk_table_attach (GTK_TABLE (table1), label_unit,       0, 1, 2, 3, GTK_FILL, GTK_FILL, PADDING, PADDING);
		gtk_table_attach (GTK_TABLE (table1), widgets.unit,     1, 2, 2, 3, GTK_FILL, GTK_FILL, PADDING, PADDING);

		// Light type Directional Options
		table2 = gtk_table_new(2, 3, FALSE);
		gtk_notebook_append_page (GTK_NOTEBOOK (widgets.light_type_sub_options), table2, NULL);

		// Angle
		widgets.angle = gtk_spin_button_new_with_range (-180, 180, 1);
		gtk_spin_button_set_digits (GTK_SPIN_BUTTON (widgets.angle), 0);
		gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.angle), 
								   atan2 (luce_options->GetDirection().y, -luce_options->GetDirection().x));
		g_signal_connect (widgets.angle,
						  "value-changed",
						  G_CALLBACK (angle_value_changed_callback),
						  luce_options);

		gtk_table_attach (GTK_TABLE (table2), label_angle,   0, 1, 0, 1, GTK_FILL, GTK_FILL, PADDING, PADDING);
		gtk_table_attach (GTK_TABLE (table2), widgets.angle, 1, 2, 0, 1, GTK_FILL, GTK_FILL, PADDING, PADDING);
		
	}

	// Algorithm
	algorithm_option_frame = gtk_frame_new (_("Algorithm Options"));
	gtk_table_attach (GTK_TABLE (options_table), algorithm_option_frame, 0, 2, 1, 2, GTK_FILL, GTK_FILL, PADDING, PADDING);
	{
		GtkWidget *hbox = gtk_hbox_new (FALSE, PADDING);
		GtkWidget *vbox1 = gtk_vbox_new (FALSE, PADDING);
		GtkWidget *vbox2 = gtk_vbox_new (FALSE, PADDING);
		GtkWidget *label_color;
		GtkWidget *label_mask;
		const GimpRGB color = {
			(gdouble)luce_options->GetZero(0),
			(gdouble)luce_options->GetZero(1),
			(gdouble)luce_options->GetZero(2)
		};

		gtk_container_add (GTK_CONTAINER (algorithm_option_frame), hbox);
		gtk_container_add (GTK_CONTAINER (hbox), vbox1);
		gtk_container_add (GTK_CONTAINER (hbox), gtk_vseparator_new());
		gtk_container_add (GTK_CONTAINER (hbox), vbox2);
		gtk_container_set_border_width (GTK_CONTAINER (hbox), PADDING);
		gtk_container_set_border_width (GTK_CONTAINER (vbox1), PADDING);
		gtk_container_set_border_width (GTK_CONTAINER (vbox2), PADDING);

		// Zero Type
		widgets.algorithm_type_zero = gtk_radio_button_new_with_label (
			NULL,
			_("Zero(Original)"));
		gtk_toggle_button_set_active (
			GTK_TOGGLE_BUTTON (widgets.algorithm_type_zero),
			luce_options->IsZeroMode());
		g_signal_connect (widgets.algorithm_type_zero, "clicked",
			G_CALLBACK (algorithm_type_zero_clicked_callback), luce_options);
		gtk_box_pack_start (GTK_BOX(vbox1), widgets.algorithm_type_zero, FALSE, FALSE, PADDING);

		// Light Form Type
		widgets.algorithm_type_light_form = gtk_radio_button_new_with_label (
			gtk_radio_button_get_group (GTK_RADIO_BUTTON (widgets.algorithm_type_zero)),
			_("Light form"));
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widgets.algorithm_type_light_form),
									  luce_options->IsLightformMode());
		g_signal_connect (widgets.algorithm_type_light_form,
						  "clicked",
						  G_CALLBACK (algorithm_type_light_form_clicked_callback),
						  luce_options);
		gtk_box_pack_start (GTK_BOX(vbox1), widgets.algorithm_type_light_form, FALSE, FALSE, PADDING);


		// Labels
		label_color = gtk_label_new (_("Color:"));
		label_mask = gtk_label_new (_("Mask:"));
		gtk_misc_set_alignment (GTK_MISC (label_color), 0.0, 0.5);
		gtk_misc_set_alignment (GTK_MISC (label_mask), 0.0, 0.5);

		// Algorithm type Options
		widgets.algorithm_type_option_table = gtk_table_new (5, 2, FALSE);
		gtk_container_add (GTK_CONTAINER (vbox2), widgets.algorithm_type_option_table);

		// Color
		widgets.color = gimp_color_button_new (_("Zero Color"), 50, 16, &color, 
											   GIMP_COLOR_AREA_FLAT);
		g_signal_connect (widgets.color,
						  "color-changed",
						  G_CALLBACK (color_color_changed_callback),
						  luce_options);

		// Mask
		widgets.mask = gtk_spin_button_new_with_range (-255, +255, 1);
		gtk_spin_button_set_digits (GTK_SPIN_BUTTON (widgets.mask), 0);
		gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.mask), 
								   (int)(luce_options->GetZero(4) * 255));
		g_signal_connect (widgets.mask,
						  "value-changed",
						  G_CALLBACK (mask_value_changed_callback),
						  luce_options);

		// if layer does not have mask, disable the mask widgets.
		{
			int mask_ID = gimp_layer_get_mask (drawable->drawable_id);
			gtk_widget_set_sensitive (label_mask,   mask_ID != -1);
			gtk_widget_set_sensitive (widgets.mask, mask_ID != -1);
		}

		// Normalize light
		widgets.normalize_light = gtk_check_button_new_with_label (_("Normalize light"));
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widgets.normalize_light),
									  luce_options->IsNormalizePositive());
		g_signal_connect (widgets.normalize_light,
						  "clicked",
						  G_CALLBACK (normalize_light_clicked_callback),
						  luce_options);

		// Normalize shadow
		widgets.normalize_shadow = gtk_check_button_new_with_label (_("Normalize shadow"));
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widgets.normalize_shadow),
									  luce_options->IsNormalizeNegative());
		g_signal_connect (widgets.normalize_shadow,
						  "clicked",
						  G_CALLBACK (normalize_shadow_clicked_callback),
						  luce_options);

		gtk_table_attach (GTK_TABLE (widgets.algorithm_type_option_table), label_color,              0, 1, 0, 1, GTK_FILL, GTK_FILL, PADDING, PADDING);
		gtk_table_attach (GTK_TABLE (widgets.algorithm_type_option_table), widgets.color,            1, 2, 0, 1, GTK_FILL, GTK_FILL, PADDING, PADDING);
		gtk_table_attach (GTK_TABLE (widgets.algorithm_type_option_table), label_mask,               0, 1, 1, 2, GTK_FILL, GTK_FILL, PADDING, PADDING);
		gtk_table_attach (GTK_TABLE (widgets.algorithm_type_option_table), widgets.mask,             1, 2, 1, 2, GTK_FILL, GTK_FILL, PADDING, PADDING);
		gtk_table_attach (GTK_TABLE (widgets.algorithm_type_option_table), widgets.normalize_light,  0, 2, 2, 3, GTK_FILL, GTK_FILL, PADDING, PADDING);
		gtk_table_attach (GTK_TABLE (widgets.algorithm_type_option_table), widgets.normalize_shadow, 0, 2, 3, 4, GTK_FILL, GTK_FILL, PADDING, PADDING);
	}

	// Scale Intensity
	scale_option_frame = gtk_frame_new (_("Scale Options"));
	gtk_table_attach (GTK_TABLE (options_table), scale_option_frame, 0, 2, 2, 3, GTK_FILL, GTK_FILL, PADDING, PADDING);
	{
		GtkWidget *scale_option_table = gtk_table_new (3, 2, FALSE);
		GtkWidget *label_light = gtk_label_new (_("Light:"));
		GtkWidget *label_shadow = gtk_label_new (_("Shadow:"));
		gtk_misc_set_alignment (GTK_MISC (label_light), 0.0, 0.5);
		gtk_misc_set_alignment (GTK_MISC (label_shadow), 0.0, 0.5);

		// Options table
		gtk_container_add (GTK_CONTAINER (scale_option_frame), scale_option_table);
		gtk_container_set_border_width (GTK_CONTAINER(scale_option_table), PADDING);

		// Light Scale
		widgets.scale_light = gtk_spin_button_new_with_range (1, 10000, 1);
		gtk_spin_button_set_digits (GTK_SPIN_BUTTON (widgets.scale_light), 2);
		gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.scale_light), 
								   INTENSITY_LUCE_VALUE_TO_SPIN_VALUE(luce_options->GetPositiveIntensity()));
		g_signal_connect (widgets.scale_light,
						  "value-changed",
						  G_CALLBACK (scale_light_value_changed_callback),
						  luce_options);

		// Shadow Scale
		widgets.scale_shadow = gtk_spin_button_new_with_range (1, 10000, 1);
		gtk_spin_button_set_digits (GTK_SPIN_BUTTON (widgets.scale_shadow), 2);
		gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.scale_shadow), 
								   INTENSITY_LUCE_VALUE_TO_SPIN_VALUE(luce_options->GetNegativeIntensity()));
		g_signal_connect (widgets.scale_shadow,
						  "value-changed",
						  G_CALLBACK (scale_shadow_value_changed_callback),
						  luce_options);

		
		// Light Scale Slider
		widgets.adjustment_light_scale = gtk_adjustment_new (2000, 0, 4000, 1, 1000, 0);
		widgets.scale_light_slider = gtk_hscale_new (GTK_ADJUSTMENT (widgets.adjustment_light_scale));
		gtk_scale_set_draw_value (GTK_SCALE (widgets.scale_light_slider), FALSE);
		gtk_widget_set_size_request (widgets.scale_light_slider, 110, -1);
		g_signal_connect (widgets.adjustment_light_scale,
						  "value-changed",
						  G_CALLBACK (adjustment_light_scale_value_changed),
						  luce_options);

		// Shadow Scale Slider
		widgets.adjustment_shadow_scale = gtk_adjustment_new (2000, 0, 4000, 1, 1000, 0);
		widgets.scale_shadow_slider = gtk_hscale_new (GTK_ADJUSTMENT (widgets.adjustment_shadow_scale));
		gtk_scale_set_draw_value (GTK_SCALE (widgets.scale_shadow_slider), FALSE);
		gtk_widget_set_size_request (widgets.scale_shadow_slider, 110, -1);
		g_signal_connect (widgets.adjustment_shadow_scale,
						  "value-changed",
						  G_CALLBACK (adjustment_shadow_scale_value_changed),
						  luce_options);

		gtk_table_attach (GTK_TABLE (scale_option_table), label_light,                 0, 1, 0, 1, GTK_FILL, GTK_FILL, PADDING*2, PADDING);
		gtk_table_attach (GTK_TABLE (scale_option_table), widgets.scale_light_slider,  1, 2, 0, 1, GTK_FILL, GTK_FILL, PADDING*2  , PADDING);
		gtk_table_attach (GTK_TABLE (scale_option_table), widgets.scale_light,         2, 3, 0, 1, GTK_FILL, GTK_FILL, PADDING*2, PADDING);
		gtk_table_attach (GTK_TABLE (scale_option_table), label_shadow,                0, 1, 1, 2, GTK_FILL, GTK_FILL, PADDING*2, PADDING);
		gtk_table_attach (GTK_TABLE (scale_option_table), widgets.scale_shadow_slider, 1, 2, 1, 2, GTK_FILL, GTK_FILL, PADDING*2,   PADDING);
		gtk_table_attach (GTK_TABLE (scale_option_table), widgets.scale_shadow,        2, 3, 1, 2, GTK_FILL, GTK_FILL, PADDING*2, PADDING);
	}

	// Source
	source_option_frame = gtk_frame_new (_("Source Options"));
	gtk_table_attach (GTK_TABLE (options_table), source_option_frame, 0, 1, 3, 4, GTK_FILL, GTK_FILL, PADDING, PADDING);
	{
		GtkWidget *vbox = gtk_vbox_new (FALSE, PADDING);

		// Options table
		gtk_container_add (GTK_CONTAINER (source_option_frame), vbox);
		gtk_container_set_border_width (GTK_CONTAINER(vbox), PADDING);

		// Add source
		widgets.source_add = gtk_check_button_new_with_label (_("Add source"));
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widgets.source_add),
									  luce_options->IsAddSource());
		g_signal_connect (widgets.source_add,
						  "clicked",
						  G_CALLBACK (source_add_clicked_callback),
						  luce_options);
		gtk_box_pack_start (GTK_BOX(vbox), widgets.source_add, FALSE, FALSE, PADDING);

		// Apply zero
		widgets.source_zero = gtk_check_button_new_with_label (_("Apply zero"));
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widgets.source_zero),
									  luce_options->IsSrcApplyZero());
		g_signal_connect (widgets.source_zero,
						  "clicked",
						  G_CALLBACK (source_zero_clicked_callback),
						  luce_options);
		gtk_box_pack_start (GTK_BOX(vbox), widgets.source_zero, FALSE, FALSE, PADDING);

		// Normalize
		widgets.source_normalize = gtk_check_button_new_with_label (_("Normalize"));
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widgets.source_normalize),
									  luce_options->IsSrcNormalizePositive());
		g_signal_connect (widgets.source_normalize,
						  "clicked",
						  G_CALLBACK (source_normalize_clicked_callback),
						  luce_options);
		gtk_box_pack_start (GTK_BOX(vbox), widgets.source_normalize, FALSE, FALSE, PADDING);
	}

	// Threads
	thread_option_frame = gtk_frame_new (_("Thread Options"));
	gtk_table_attach (GTK_TABLE (options_table), thread_option_frame, 1, 2, 3, 4, GTK_FILL, GTK_FILL, PADDING, PADDING);
	{
		GtkWidget *table = gtk_table_new (1, 2, FALSE);
		GtkWidget *label_thread = gtk_label_new (_("Threads:"));
		int i;
		int processor_number = get_processor_number ();

		gtk_misc_set_alignment (GTK_MISC (label_thread), 0.0, 0.5);

		// Options table
		gtk_container_add (GTK_CONTAINER (thread_option_frame), table);
		gtk_container_set_border_width (GTK_CONTAINER(table), PADDING);

		// Thread number
		widgets.thread_number = gimp_int_combo_box_new (" 1", 1, NULL);
		for (i = 2; i <= processor_number; i++) {
			char label[4];
			snprintf (label, 4, "%2d\0", i);
			gimp_int_combo_box_append (GIMP_INT_COMBO_BOX (widgets.thread_number),
										GIMP_INT_STORE_VALUE,    i, 
										GIMP_INT_STORE_LABEL,    label, 
										GIMP_INT_STORE_STOCK_ID, NULL,
										-1); 
		}
		gimp_int_combo_box_connect (GIMP_INT_COMBO_BOX (widgets.thread_number), 
									luce_options->GetNumThreadsToUse(),
									G_CALLBACK(thread_number_changed_callback),
									luce_options);
		
		gtk_table_attach (GTK_TABLE (table), label_thread,          0, 1, 0, 1, GTK_FILL, GTK_FILL, PADDING, PADDING);
		gtk_table_attach (GTK_TABLE (table), widgets.thread_number, 1, 2, 0, 1, GTK_FILL, GTK_FILL, PADDING, PADDING);
	}

	// Show dialog
	widgets.dialog_result = ERROR_NOTHING;
	gtk_window_set_position (GTK_WINDOW (widgets.dialog), GTK_WIN_POS_CENTER_ALWAYS);
	gtk_widget_show_all (widgets.dialog);

	// Set options
	if (luce_options->IsPointLight()) {
		light_type_point_clicked_callback (widgets.light_type_point, luce_options);
	} else {
		light_type_directional_clicked_callback (widgets.light_type_directional, luce_options);
	}
	if (luce_options->IsZeroMode()) {
		algorithm_type_zero_clicked_callback (widgets.algorithm_type_zero, luce_options);
	} else {
		algorithm_type_light_form_clicked_callback (widgets.algorithm_type_light_form, luce_options);
	}
	scale_light_value_changed_callback (widgets.scale_light, luce_options);
	scale_shadow_value_changed_callback (widgets.scale_shadow, luce_options);
	
	update_source_sensitive ();

	luce_gimp_preview_enable_update (TRUE);
	luce_gimp_preview_update_buffer (luce_options);

	gtk_main ();

	return widgets.dialog_result;
#undef PADDING
}
