/*
 * (C) 2002 Antonino Perricone <don_perricone@libero.it>
 * (C) 2013 Ohtsuka Yoshio <ohtsuka.yoshio@gmail.com>
 *
 * This work is licensed under a Creative Commons 
 * Attribution-NonCommercial-ShareAlike 3.0 Unported License.
 * See this link for more informations: 
 * http://creativecommons.org/licenses/by-nc-sa/3.0/
 */
#include "luce_gimp_main.h"
#include "luce_gimp_dialog.h"

typedef struct _PluginVals {
	gboolean   point_light;
	gboolean   quadratic_attenuation;
	gboolean   zero_mode;
	gboolean   add_source;
	gboolean   source_zero;
	gboolean   source_positive_normalize;
	gboolean   source_negative_normalize;
	gboolean   positive_normalize;
	gboolean   negative_normalize;
	gint32     num_threads_to_use;
	gdouble    positive_intensity;
	gdouble    negative_intensity;
	gdouble    center_x;
	gdouble    center_y;
	gdouble    direction_x;
	gdouble    direction_y;
	gdouble    zeros[5];
} PluginVals;


static PluginError	luce_gimp_run							(GimpDrawable		*drawable,
															 LuceOptions		*luce_options,
															 gdouble			progress_st,
															 gdouble			progress_ed);
static void			luce_gimp_vals_to_luce_options			(LuceOptions		*luce_options,
															 PluginVals			*vals);
static void			luce_gimp_luce_options_to_vals			(LuceOptions		*luce_options,
															 PluginVals			*vals);
static void			luce_gimp_gimpparams_to_luce_options	(LuceOptions		*luce_options,
															 const GimpParam	*param);

/*static void init    (void);*/
/*static void quit    (void);*/
static void query	(void);
static void run		(const gchar		*name,
					 gint				nparams,
					 const GimpParam	*param,
					 gint				*nreturn_vals,
					 GimpParam			**return_vals);

GimpPlugInInfo PLUG_IN_INFO = {
	NULL,  /* init  */
	NULL,  /* quit  */
	query, /* query */
	run    /* run   */
};

MAIN ()

static void
query (void)
{
	static const GimpParamDef args[] = {
	{ GIMP_PDB_INT32,   (gchar*)"run-mode",     (gchar*)"Interactive=0, Non-Interactive=1" },
	{ GIMP_PDB_IMAGE,   (gchar*)"image",        (gchar*)"Input image" },
	{ GIMP_PDB_DRAWABLE,(gchar*)"drawable",     (gchar*)"Input drawable" },
	{ GIMP_PDB_INT32,   (gchar*)"point light",               (gchar*)"TRUE=1, FALSE=0" },
	{ GIMP_PDB_INT32,   (gchar*)"quadratic attenuation",     (gchar*)"TRUE=1, FALSE=0" },
	{ GIMP_PDB_INT32,   (gchar*)"zero mode",                 (gchar*)"TRUE=1, FALSE=0" },
	{ GIMP_PDB_INT32,   (gchar*)"add source",                (gchar*)"TRUE=1, FALSE=0" },
	{ GIMP_PDB_INT32,   (gchar*)"source zero",               (gchar*)"TRUE=1, FALSE=0" },
	{ GIMP_PDB_INT32,   (gchar*)"source positive normalize", (gchar*)"TRUE=1, FALSE=0" },
	{ GIMP_PDB_INT32,   (gchar*)"source negative normalize", (gchar*)"TRUE=1, FALSE=0" },
	{ GIMP_PDB_INT32,   (gchar*)"positive normalize",        (gchar*)"TRUE=1, FALSE=0" },
	{ GIMP_PDB_INT32,   (gchar*)"negative normalize",        (gchar*)"TRUE=1, FALSE=0" },
	{ GIMP_PDB_INT32,   (gchar*)"num threads to use",        (gchar*)"1" },
	{ GIMP_PDB_FLOAT,   (gchar*)"positive intensity",        (gchar*)"1.0f" },
	{ GIMP_PDB_FLOAT,   (gchar*)"negative intensity",        (gchar*)"1.0f" },
	{ GIMP_PDB_FLOAT,   (gchar*)"center x",                  (gchar*)"center x for point light (0.5) not pixel but percent" },
	{ GIMP_PDB_FLOAT,   (gchar*)"center y",                  (gchar*)"center y for point light (0.5) not pixel but percent" },
	{ GIMP_PDB_FLOAT,   (gchar*)"direction x",               (gchar*)"cos for direction light (-0.707f)" },
	{ GIMP_PDB_FLOAT,   (gchar*)"direction y",               (gchar*)"sin for direction light (-0.707f)" },
	{ GIMP_PDB_FLOAT,   (gchar*)"zeros0",                    (gchar*)"zero color red plane (0.0f)"},
	{ GIMP_PDB_FLOAT,   (gchar*)"zeros1",                    (gchar*)"zero color green plane (0.0f)"},
	{ GIMP_PDB_FLOAT,   (gchar*)"zeros2",                    (gchar*)"zero color blue plane (0.0f)"},
	{ GIMP_PDB_FLOAT,   (gchar*)"zeros3",                    (gchar*)"zero color alpha plane (0.0f)"},
	{ GIMP_PDB_FLOAT,   (gchar*)"zeros4",                    (gchar*)"zero color mask plane (0.0f)"},
	};

	gimp_install_procedure (PLUGIN_ID,
							_(PLUGIN_DESCRIPTION),
							_(PLUGIN_DESCRIPTION),
							PLUGIN_AUTHOR,
							PLUGIN_AUTHOR,
							PLUGIN_DATE,
							_("Luce..."),
							"RGB*, GRAY*",
							GIMP_PLUGIN,
							G_N_ELEMENTS (args), 0,
							args, NULL);

	gimp_plugin_menu_register (PLUGIN_ID, "<Image>/Filters/Light and Shadow/Light");
}

//Start Point of Plugin
static void
run (const gchar	*name,
	gint			nparams,
	const GimpParam	*param,
	gint			*nreturn_vals,
	GimpParam		**return_vals)
{
	static GimpParam	values[1];
	GimpDrawable		*drawable;
	GimpRunMode			run_mode;
	GimpPDBStatusType	status = GIMP_PDB_SUCCESS;

	PluginVals vals;
	LuceOptions luce_options;

	INIT_I18N();

	run_mode = (GimpRunMode)param[0].data.d_int32;
	drawable = gimp_drawable_get (param[2].data.d_drawable);
	gimp_tile_cache_ntiles (2 * (drawable->width / gimp_tile_width () + 1));

	luce_options.InitDefault();

	switch (run_mode) {
		case GIMP_RUN_NONINTERACTIVE:
			if (nparams != 24) {
				status = GIMP_PDB_CALLING_ERROR;
			} else {
				luce_gimp_gimpparams_to_luce_options (&luce_options, param);
				if (luce_gimp_run (drawable, &luce_options, 0.0, 1.0) != ERROR_NOTHING) {
					status = GIMP_PDB_EXECUTION_ERROR;
					break;
				}
			}
			break;

		case GIMP_RUN_INTERACTIVE:
			if (gimp_get_data (PLUGIN_ID, &vals)) {
				luce_gimp_vals_to_luce_options (&luce_options, &vals);
			}
			if (luce_gimp_dialog_run (drawable, &luce_options) != ERROR_NOTHING) {
				status =GIMP_PDB_CANCEL;
				break;
			}
			if (luce_gimp_run (drawable, &luce_options, 0.0, 1.0) != ERROR_NOTHING) {
				status = GIMP_PDB_EXECUTION_ERROR;
				break;
			}
			luce_gimp_luce_options_to_vals(&luce_options, &vals);
			gimp_set_data (PLUGIN_ID, &vals, sizeof(PluginVals));
			gimp_displays_flush ();
			break;

		case GIMP_RUN_WITH_LAST_VALS:
			if (gimp_get_data (PLUGIN_ID, &vals)) {
				luce_gimp_vals_to_luce_options (&luce_options, &vals);
			}
			if (luce_gimp_run (drawable, &luce_options, 0.0, 1.0) != ERROR_NOTHING) {
				status = GIMP_PDB_EXECUTION_ERROR;
				break;
			}
			gimp_displays_flush ();
			break;
	}

	gimp_drawable_detach (drawable);

	*nreturn_vals = 1;
	*return_vals = values;
	values[0].type = GIMP_PDB_STATUS;
	values[0].data.d_status = status;
	return;
}

PluginError
luce_gimp_run (GimpDrawable	*drawable,
			   LuceOptions	*luce_options,
			   gdouble		progress_st,
			   gdouble		progress_ed)
{
	CLuce			luce;
	GimpPixelRgn	src_region, dst_region;
	guchar			*src_buf, *dst_buf;
	guint			width, height, bpp, stride;
	gboolean		has_alpha, has_mask, is_mask;
	gint32			mask_ID;
	gdouble			progress_value;

	width = drawable->width;
	height = drawable->height;
	bpp = drawable->bpp;
	stride = width * bpp;
	has_alpha = gimp_drawable_has_alpha (drawable->drawable_id);
	has_mask = FALSE;
	if (gimp_drawable_is_layer(drawable->drawable_id)) {
		mask_ID = gimp_layer_get_mask (drawable->drawable_id);
		if (mask_ID != -1) {
			has_mask = TRUE;
		}
	}
	is_mask = gimp_drawable_is_layer_mask (drawable->drawable_id);

	progress_value = progress_st;
	if (progress_st == 0) {
		gimp_progress_init(_("Rendering Luce..."));
	}
	gimp_progress_update (progress_value);

	src_buf = g_new (guchar, stride * height);
	dst_buf = g_new (guchar, stride * height);
	gimp_pixel_rgn_init (&src_region, drawable, 0, 0, width, height, FALSE, FALSE);
	gimp_pixel_rgn_init (&dst_region, drawable, 0, 0, width, height, TRUE, TRUE);
	gimp_pixel_rgn_get_rect (&src_region, src_buf, 0, 0, width, height);

	luce.width = width;
	luce.height =height;

	luce.input.iPixelSize = bpp;
	luce.input.iStride = stride;
	luce.input.pBits = src_buf;

	luce.output.iPixelSize = bpp;
	luce.output.iStride = stride;
	luce.output.pBits = dst_buf;

	if (has_alpha) {
		luce.alpha.pBits = luce.input.pBits + 3;
		luce.alpha.iStride = luce.input.iStride;
		luce.alpha.iPixelSize = luce.input.iPixelSize;
	} else {
		luce.alpha.pBits = NULL;
	}

	luce.pOpt = luce_options;

	for (int i = 0; i < luce.input.iPixelSize; i++) {
		if (has_alpha && i == luce.input.iPixelSize -1) {
			luce.alpha.pBits = NULL;
		}
		luce.currentZero = i;
		luce.Do8Bit();
		luce.input.pBits++;
		luce.output.pBits++;

		progress_value = (i+1.0) / (luce.input.iPixelSize + (has_mask ? 1 : 0));
		gimp_progress_update (progress_value);
	}

	gimp_pixel_rgn_set_rect (&dst_region, dst_buf, 0, 0, width, height);
	g_free (src_buf);
	g_free (dst_buf);
	
	if (has_mask && !is_mask) {
		PluginError error;
		error = luce_gimp_run (gimp_drawable_get (mask_ID), luce_options, progress_value, 1);
		if (error != ERROR_NOTHING) {
			return error;
		}
	}

	gimp_progress_update (progress_ed);
	gimp_drawable_flush (drawable);
	gimp_drawable_merge_shadow (drawable->drawable_id, TRUE);
	gimp_drawable_update (drawable->drawable_id, 0, 0, width, height);

	gimp_displays_flush ();

	return ERROR_NOTHING;
}

static void
luce_gimp_luce_options_to_vals (LuceOptions	*luce_options,
								PluginVals	*vals)
{
	vals->point_light               = luce_options->IsPointLight();
	vals->quadratic_attenuation     = luce_options->IsQuadraticAttenuation();
	vals->zero_mode                 = luce_options->IsZeroMode();
	vals->add_source                = luce_options->IsAddSource();
	vals->source_zero               = luce_options->IsSrcApplyZero();
	vals->source_positive_normalize = luce_options->IsSrcNormalizePositive();
	vals->source_negative_normalize = luce_options->IsSrcNormalizeNegative();
	vals->positive_normalize        = luce_options->IsNormalizePositive();
	vals->negative_normalize        = luce_options->IsNormalizeNegative();
	vals->num_threads_to_use        = luce_options->GetNumThreadsToUse();
	vals->positive_intensity        = luce_options->GetPositiveIntensity();
	vals->negative_intensity        = luce_options->GetNegativeIntensity();
	vals->center_x                  = luce_options->GetCenter().x;
	vals->center_y                  = luce_options->GetCenter().y;
	vals->direction_x               = luce_options->GetDirection().x;
	vals->direction_y               = luce_options->GetDirection().y;
	vals->zeros[0]                  = luce_options->GetZero(0);
	vals->zeros[1]                  = luce_options->GetZero(1);
	vals->zeros[2]                  = luce_options->GetZero(2);
	vals->zeros[3]                  = luce_options->GetZero(3);
	vals->zeros[4]                  = luce_options->GetZero(4);
}

static void
luce_gimp_vals_to_luce_options (LuceOptions *luce_options,
								PluginVals *vals) {
	if (vals->point_light) {
		luce_options->SetPointLight();
	} else {
		luce_options->SetDirectionaLight();
	}
	if (vals->quadratic_attenuation) {
		luce_options->SetQuadraticAttenuation();
	} else {
		luce_options->SetLinearAttenuation();
	}
	if (vals->zero_mode) {
		luce_options->SetZeroMode();
	} else {
		luce_options->SetLightformMode();
	}
	if (vals->add_source) {
		luce_options->SetAddSource();
	} else {
		luce_options->SetNotAddSource();
	}
	if (vals->source_zero) {
		luce_options->SetSrcApplyZero();
	} else {
		luce_options->SetSrcIgnoreZero();
	}
	if (vals->source_positive_normalize) {
		luce_options->SetSrcNormalizePositive();
	} else {
		luce_options->SetSrcMantainPositiveScale();
	}
	if (vals->source_negative_normalize) {
		luce_options->SetSrcNormalizeNegative();
	} else {
		luce_options->SetSrcMantainNegativeScale();
	}
	if (vals->positive_normalize) {
		luce_options->SetNormalizePositive();
	} else {
		luce_options->SetMantainPositiveScale();
	}
	if (vals->negative_normalize) {
		luce_options->SetNormalizeNegative();
	} else {
		luce_options->SetMantainNegativeScale();
	}
	luce_options->SetNumThreadsToUse(vals->num_threads_to_use);
	luce_options->SetPositiveIntensity((float)vals->positive_intensity);
	luce_options->SetNegativeIntensity((float)vals->negative_intensity);
	luce_options->SetCenter(vals->center_x, vals->center_y);
	luce_options->SetDirection((float)vals->direction_x, (float)vals->direction_y);
	luce_options->SetZero(0, (float)vals->zeros[0]);
	luce_options->SetZero(1, (float)vals->zeros[1]);
	luce_options->SetZero(2, (float)vals->zeros[2]);
	luce_options->SetZero(3, (float)vals->zeros[3]);
	luce_options->SetZero(4, (float)vals->zeros[4]);
}

static void
luce_gimp_gimpparams_to_luce_options (LuceOptions		*luce_options,
									 const GimpParam	*param)
{
	if (param[3].data.d_int32 == 1) {
		luce_options->SetPointLight();
	} else {
		luce_options->SetDirectionaLight();
	}
	if (param[4].data.d_int32 == 1) {
		luce_options->SetQuadraticAttenuation();
	} else {
		luce_options->SetLinearAttenuation();
	}
	if (param[5].data.d_int32 == 1) {
		luce_options->SetZeroMode();
	} else {
		luce_options->SetLightformMode();
	}
	if (param[6].data.d_int32 == 1) {
		luce_options->SetAddSource();
	} else {
		luce_options->SetNotAddSource();
	}
	if (param[7].data.d_int32 == 1) {
		luce_options->SetSrcApplyZero();
	} else {
		luce_options->SetSrcIgnoreZero();
	}
	if (param[8].data.d_int32 == 1) {
		luce_options->SetSrcNormalizePositive();
	} else {
		luce_options->SetSrcMantainPositiveScale();
	}
	if (param[9].data.d_int32 == 1) {
		luce_options->SetSrcNormalizeNegative();
	} else {
		luce_options->SetSrcMantainNegativeScale();
	}
	if (param[10].data.d_int32 == 1) {
		luce_options->SetNormalizePositive();
	} else {
		luce_options->SetMantainPositiveScale();
	}
	if (param[11].data.d_int32 == 1) {
		luce_options->SetNormalizeNegative();
	} else {
		luce_options->SetMantainNegativeScale();
	}
	luce_options->SetNumThreadsToUse(param[12].data.d_int32);
	luce_options->SetPositiveIntensity((float)param[13].data.d_float);
	luce_options->SetNegativeIntensity((float)param[14].data.d_float);
	luce_options->SetCenter(param[15].data.d_float, param[16].data.d_float);
	luce_options->SetDirection((float)param[17].data.d_float, (float)param[18].data.d_float);
	luce_options->SetZero(0, (float)param[19].data.d_float);
	luce_options->SetZero(1, (float)param[20].data.d_float);
	luce_options->SetZero(2, (float)param[21].data.d_float);
	luce_options->SetZero(3, (float)param[22].data.d_float);
	luce_options->SetZero(4, (float)param[23].data.d_float);
}
