/*
 * (C) 2002 Antonino Perricone <don_perricone@libero.it>
 * (C) 2013 Ohtsuka Yoshio <ohtsuka.yoshio@gmail.com>
 *
 * This work is licensed under a Creative Commons 
 * Attribution-NonCommercial-ShareAlike 3.0 Unported License.
 * See this link for more informations: 
 * http://creativecommons.org/licenses/by-nc-sa/3.0/
 */
#include <glib.h>
#include <gtk/gtk.h>
#include "luce_gimp_main.h"
#include "luce_gimp_dialog.h"
#include "luce_gimp_preview.h"

static GtkWidget *preview = NULL;
static GdkPixbuf *src = NULL;
static GdkPixbuf *dst = NULL;
static gboolean	 enable_update_preview_buffer = FALSE;

#define PREVIEW_W 400
#define PREVIEW_H 400

static void		init_preview_buf			(GimpDrawable	*drawable);
static void		update_preview_buf			(LuceOptions	*luce_options);
static void		clear_buf					();
static void		get_origin_point			(gint			*x,
											 gint			*y);
static void		preview_realize_callback	(GtkWidget		*widget,
											 LuceOptions	*luce_options);
static gboolean	preview_event_callback		(GtkWidget		*widget,
											 GdkEvent		*event, 
											 LuceOptions	*luce_options);
static gboolean	preview_expose_callback		(GtkWidget		*widget,
											 GdkEventExpose	*event,
											 LuceOptions	*luce_options);
static void		preview_destroy_callback	(GtkWidget		*widget,
											 LuceOptions	*luce_options);
GtkWidget*		luce_gimp_preview_new		(LuceOptions	*luce_options,
											 GimpDrawable	*drawable);

static void
init_preview_buf (GimpDrawable *drawable)
{
	gint preview_w, preview_h;
	gint image_w, image_h;
	gint buf_w, buf_h;

	if (src != NULL || dst != NULL) {
		clear_buf();
	}
	
	preview_w = PREVIEW_W;
	preview_h = PREVIEW_H;
	image_w = drawable->width;
	image_h = drawable->height;

	if (image_w > image_h) {
		if (preview_w > image_w) {
			buf_w = image_w;
			buf_h = image_h;
		} else {
			buf_w = preview_w;
			buf_h = (gint)((image_h / (gdouble)image_w) * preview_w);
		}
	} else {
		if (preview_h > image_h) {
			buf_w = image_w;
			buf_h = image_h;
		} else {
			buf_w = (gint)((image_w / (gdouble)image_h) * preview_h); 
			buf_h = preview_h;
		}
	}

	src = gimp_drawable_get_thumbnail (drawable->drawable_id,
									   buf_w,
									   buf_h,
									   GIMP_PIXBUF_KEEP_ALPHA);

	dst = gdk_pixbuf_new (GDK_COLORSPACE_RGB,
						  gdk_pixbuf_get_has_alpha (src),
						  8,
						  buf_w,
						  buf_h);
}

static void
update_preview_buf (LuceOptions *luce_options)
{
	CLuce luce;

	if (enable_update_preview_buffer == FALSE) {
		return;
	}

	if (preview == NULL || src == NULL || dst == NULL) return;
	

	gboolean has_alpha = gdk_pixbuf_get_has_alpha(src);

	luce.width = gdk_pixbuf_get_width (src);
	luce.height = gdk_pixbuf_get_height (src);
	luce.input.iPixelSize = gdk_pixbuf_get_n_channels (src);
	luce.input.iStride = gdk_pixbuf_get_rowstride (src);
	luce.input.pBits = gdk_pixbuf_get_pixels (src);
	luce.output.iPixelSize = gdk_pixbuf_get_n_channels (dst);
	luce.output.iStride = gdk_pixbuf_get_rowstride (dst);
	luce.output.pBits = gdk_pixbuf_get_pixels (dst);
	luce.pOpt = luce_options;

	if (has_alpha) {
		luce.alpha.pBits = luce.input.pBits + 3;
		luce.alpha.iStride = luce.input.iStride;
		luce.alpha.iPixelSize = luce.input.iPixelSize;
	} else {
		luce.alpha.pBits = NULL;
	}

	for (gint i = 0; i < luce.input.iPixelSize; i++) {
		if (has_alpha && i == luce.input.iPixelSize -1) {
			luce.alpha.pBits = NULL;
		}
		luce.currentZero = i;
		luce.Do8Bit();
		luce.input.pBits++;
		luce.output.pBits++;
	}

	if (gdk_pixbuf_get_has_alpha (dst)) {
		GdkPixbuf *tmp;
		tmp = gdk_pixbuf_composite_color_simple (dst,
												 luce.width,
												 luce.height,
												 GDK_INTERP_NEAREST,
												 255,
												 GIMP_CHECK_SIZE,
												 0x66666666,
												 0x99999999);
		g_object_unref (dst);
		dst = tmp;
	}

}

static void
clear_buf () 
{
	if (src != NULL) {
		g_object_unref (src);
		src = NULL;
	}
	if (dst != NULL) {
		g_object_unref (dst);
		dst = NULL;
	}
}

static void
get_origin_point (gint *x, gint *y)
{
	GtkAllocation alloc;
	gtk_widget_get_allocation (preview, &alloc);
	*x = (alloc.width - gdk_pixbuf_get_width (dst)) / 2;
	*y = (alloc.height - gdk_pixbuf_get_height (dst)) / 2;
}

static void
preview_realize_callback (GtkWidget		*widget,
						  LuceOptions	*luce_options)
{
	gdk_window_set_cursor (preview->window, gdk_cursor_new (GDK_ARROW));

	GdkDisplay *display = gtk_widget_get_display (widget);

	gdk_window_set_cursor (gtk_widget_get_window (widget),
						   gdk_cursor_new_for_display (display, GDK_CROSSHAIR));
}

static gboolean
preview_event_callback (GtkWidget		*widget,
						GdkEvent		*event,
						LuceOptions		*luce_options)
{
	static gboolean pressed = FALSE;
	gint origx, origy, bufw, bufh;
	bufw = gdk_pixbuf_get_width (src);
	bufh = gdk_pixbuf_get_height (src);
	get_origin_point (&origx, &origy);

	switch (event->type) {
		case GDK_EXPOSE:
			break;

		case GDK_BUTTON_PRESS:
			luce_gimp_dialog_set_position (luce_options, 
										   (event->button.x - origx) / bufw,
										   (event->button.y - origy) / bufh);
			preview_expose_callback (preview, &event->expose, luce_options);
			pressed = TRUE;
			break;

		case GDK_BUTTON_RELEASE:
			pressed = FALSE;
			break;

		case GDK_MOTION_NOTIFY:
			if (pressed) {
				luce_gimp_dialog_set_position (luce_options, 
											   (event->button.x - origx) / bufw,
											   (event->button.y - origy) / bufh);
				preview_expose_callback (preview, &event->expose, luce_options);
			}
			break;

	}
	return FALSE;
}

static gboolean
preview_expose_callback (GtkWidget		*widget,
						 GdkEventExpose	*event,
						 LuceOptions	*luce_options)
{
	cairo_t *cr;
	gint x=0, y=0;
	if (dst == NULL) return FALSE;

	get_origin_point (&x, &y);

	cr = gdk_cairo_create (event->window);
	gdk_cairo_set_source_pixbuf (cr, dst, x, y);
	cairo_paint (cr);
	cairo_destroy (cr);

	return FALSE;
}

static void
preview_destroy_callback (GtkWidget		*widget,
						  LuceOptions	*luce_options)
{
	clear_buf ();
}

GtkWidget*
luce_gimp_preview_new (LuceOptions	*luce_options,
					   GimpDrawable	*drawable)
{
	GdkColor gray = {0, 17767, 17767, 17767};

	preview = gtk_drawing_area_new();
	gtk_widget_add_events (preview, GDK_EXPOSURE_MASK |
                                    GDK_BUTTON_PRESS_MASK |
									GDK_BUTTON_RELEASE_MASK |
									GDK_BUTTON_MOTION_MASK);

	g_signal_connect (G_OBJECT (preview), "realize",      G_CALLBACK (preview_realize_callback), luce_options);
	g_signal_connect (G_OBJECT (preview), "event",        G_CALLBACK (preview_event_callback), luce_options);
	g_signal_connect (G_OBJECT (preview), "expose-event", G_CALLBACK (preview_expose_callback), luce_options);
	g_signal_connect (G_OBJECT (preview), "destroy",      G_CALLBACK (preview_destroy_callback), luce_options);

	gtk_widget_set_size_request (preview, PREVIEW_W, PREVIEW_H);
	gtk_widget_modify_bg (preview, GTK_STATE_NORMAL, &gray);

	init_preview_buf (drawable);

	return preview;
}

void
luce_gimp_preview_update_buffer (LuceOptions *luce_options)
{
	update_preview_buf (luce_options);
}

void
luce_gimp_preview_enable_update (gboolean update)
{
	enable_update_preview_buffer = update;
}
